#!/usr/bin/env python
# encoding: utf-8
#
# Copyright (c) 2015, Canonical Ltd.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""A fake git-like command with Guacamole."""

from __future__ import absolute_import, print_function, unicode_literals

from guacamole import Command


class GitCommit(Command):

    """The 'git commit' command."""

    def invoked(self, ctx):
        """Method called when the command is invoked."""
        print("fake git commit")

    def register_arguments(self, parser):
        """
        Guacamole method used by the argparse ingredient.

        :param parser:
            Argument parser (from :mod:`argparse`) specific to this command.
        """
        parser.add_argument(
            '-m', '--message', metavar='<msg>',
            help="Use given <msg> as the commit message")


class GitLog(Command):

    """The 'git log' command."""

    def invoked(self, ctx):
        """Method called when the command is invoked."""
        print("fake git log")

    def register_arguments(self, parser):
        """
        Guacamole method used by the argparse ingredient.

        :param parser:
            Argument parser (from :mod:`argparse`) specific to this command.
        """
        parser.add_argument(
            '--follow', action='store_true', help=(
                'Continue listing the history of a file beyond renames'
                ' (works only for a single file).'))


class Git(Command):

    """A fake git-like command."""

    sub_commands = (
        ('commit', GitCommit),
        ('log', GitLog)
    )


if __name__ == '__main__':
    Git().main()



# encoding: utf-8
# This file is part of Guacamole.
#
# Copyright 2012-2015 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Guacamole is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3,
# as published by the Free Software Foundation.
#
# Guacamole is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Guacamole.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the ansi module."""

from __future__ import absolute_import, print_function, unicode_literals

import io
import sys
import unittest

if sys.version_info[:2] <= (3, 3):
    import mock
else:
    from unittest import mock

from guacamole.ingredients.ansi import ANSIFormatter


class ANSIFormatterTests(unittest.TestCase):

    """Tests for the ANSIFormatter class."""

    def test_flush_works(self):
        """check that aprint(..., flush=True) works okay."""
        # https://github.com/zyga/guacamole/issues/9
        # This should print to our stream
        stream = io.StringIO()
        fmt = ANSIFormatter(enabled=False)
        fmt.aprint("hello world", file=stream, flush=True)
        self.assertEqual(stream.getvalue(), "hello world\n")
        # This should print to sys.stdout
        with mock.patch('sys.stdout', spec_set=True) as mocked_stdout:
            fmt.aprint("goodbye world", file=None, flush=True)
            mocked_stdout.write.assert_has_calls([
                mock.call("goodbye world"),
                mock.call("\n"),
            ])



# encoding: utf-8
# This file is part of Guacamole.
#
# Copyright 2012-2015 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Guacamole is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3,
# as published by the Free Software Foundation.
#
# Guacamole is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Guacamole.  If not, see <http://www.gnu.org/licenses/>.

"""
Ingredients for using arparse for parsing command line arguments.

This module contains two ingredients. The main one is the
:class:`ParserIngredient`. It is responsible for handling all of the command
line parsing and command argument registration. It is a part of the recipe
for the command class. Note that command dispatch is not handled by this
ingredient (see :class:`~guacamole.ingredients.cmdtree.CommandTreeIngredient`).

The second ingredient is :class:`AutocompleteIngredient` which relies on the
third-party argcomplete module to add support for automatic command line
completion to supported shells (bash).
"""

from __future__ import absolute_import, print_function, unicode_literals

import argparse

from guacamole.core import Ingredient
from guacamole.recipes import RecipeError
# from guacamole._argparse import LegacyHelpFormatter


class ParserIngredient(Ingredient):

    """
    Ingredient for using argparse to parse command line arguments.

    This ingredient uses the following Ingredient methods:

     - ``build_early_parser()``
     - ``preparse()``
     - ``build_parser()``
     - ``parse()``

    The main parser is constructed in, unsurprisingly, the
    :meth:`build_parser()` method and stored in the context as ``parser``.
    Other ingredients can be added *after* the ``ParserIngredient`` and can
    extend the available arguments (on the root parser) by using standard
    argparse APIs such as ``parser.add_argument()`` or
    ``parser.add_argument_group()``. This parser is used to handle all of
    command line in the :meth:`parse()` method.

    While most users won't concern themselves with this design decision, there
    is also a second parser, called the *early parser*, that is used to
    *pre-parse* the command line arguments. This can be used as a way
    to optimize subsequent actions as, perhaps, knowing which commands are
    going to be invoked there will be no need to instantiate and prepare *all*
    of the commands in the command tree.

    Currently this feature is not used. To take advantage of this knowledge you
    can look at the ``context.early_args`` object which contains the result of
    parsing the command line with the *early parser*. The early parser is a
    simple parser consisting of ``--help``, ``--version`` (if applicable) and
    *rest*. The *rest* argument can be used as a hint as to what is coming next
    (e.g. if it matches a name of a command we know to exist)

    After parsing is done the results of parsing the command line are stored in
    the ``context.args`` attribute. This is commonly accessed by individual
    commands from their ``invoke()`` methods.
    """

    def build_early_parser(self, context):
        """
        Create the early argument parser.

        This method creates the early argparse argument parser. The early
        parser doesn't know about any of the sub-commands so it can be used
        much earlier during the start-up process (before commands
        are loaded and initialized).
        """
        context.early_parser = self._create_early_parser(context)

    def preparse(self, context):
        """
        Parse a portion of command line arguments with the early parser.

        This method relies on ``context.argv`` and ``context.early_parser``
        and produces ``context.early_args``.

        The ``context.early_args`` object is the return value from argparse.
        It is the dict/object like namespace object.
        """
        context.early_args, unused = (
            context.early_parser.parse_known_args(context.argv))

    def build_parser(self, context):
        """
        Create the final argument parser.

        This method creates the non-early (full) argparse argument parser.
        Unlike the early counterpart it is expected to have knowledge of
        the full command tree.

        This method relies on ``context.cmd_tree`` and produces
        ``context.parser``. Other ingredients can interact with the parser
        up until :meth:`parse()` is called.
        """
        context.parser, context.max_level = self._create_parser(context)

    def parse(self, context):
        """
        Parse command line arguments.

        This method relies on ``context.argv`` and ``context.early_parser``
        and produces ``context.args``. Note that ``.argv`` is modified by
        :meth:`preparse()` so it actually has _less_ things in it.

        The ``context.args`` object is the return value from argparse.
        It is the dict/object like namespace object.
        """
        context.args = context.parser.parse_args(context.argv)

    def _create_parser(self, context):
        cmd_name, cmd_obj, cmd_subcmds = context.cmd_tree
        parser = argparse.ArgumentParser(
            prog=cmd_name, **self._get_parser_kwargs(cmd_obj))
        parser.add_argument("-h", "--help", action="help")
        self._maybe_add_version(parser, cmd_obj)
        max_level = self._add_command_to_parser(
            parser, cmd_name, cmd_obj, cmd_subcmds)
        return parser, max_level

    def _create_early_parser(self, context):
        early_parser = argparse.ArgumentParser(add_help=False)
        early_parser.add_argument(
            "rest", nargs="...", help=argparse.SUPPRESS)
        early_parser.add_argument(
            "-h", "--help", action="store_const", const=None)
        cmd_name, cmd_obj, cmd_subcmds = context.cmd_tree
        version = cmd_obj.get_cmd_version()
        if version is not None:
            early_parser.add_argument(
                "--version", action="store_const", const=None)
        return early_parser

    def _maybe_add_version(self, parser, command):
        version = command.get_cmd_version()
        if version is not None:
            # NOTE: help= is provided explicitly as argparse doesn't wrap
            # everything with _() correctly (depending on version)
            parser.add_argument(
                "--version", action="version", version=version,
                help="show program's version number and exit")

    def _get_parser_kwargs(self, command):
        return {
            'usage': command.get_cmd_usage(),
            'description': command.get_cmd_description(),
            'epilog': command.get_cmd_epilog(),
            'add_help': False,
            # formatter_class=LegacyHelpFormatter,
        }

    def _add_command_to_parser(
            self, parser, cmd_name, cmd_obj, cmd_subcmds, level=0
    ):
        # Register this command
        cmd_obj.register_arguments(parser)
        parser.set_defaults(**{'command{}'.format(level): cmd_obj})
        # Register sub-commands of this command (recursively)
        if not cmd_subcmds:
            return level
        subparsers = parser.add_subparsers(
            help="sub-command to pick")
        max_level = level
        for subcmd_name, subcmd_obj, subcmd_cmds in cmd_subcmds:
            sub_parser = subparsers.add_parser(
                subcmd_name, help=subcmd_obj.get_cmd_help(),
                **self._get_parser_kwargs(subcmd_obj))
            sub_parser.add_argument("-h", "--help", action="help")
            max_level = max(
                max_level, self._add_command_to_parser(
                    sub_parser, subcmd_name, subcmd_obj, subcmd_cmds,
                    level + 1))
        return max_level


class AutocompleteIngredient(Ingredient):

    """
    Ingredient for adding shell auto-completion.

    .. warning::
        This component is not widely tested due to difficulty of providing
        actual integration. It might be totally broken.

    .. note::
        To effectively get tab completion you need to have the ``argcomplete``
        package installed. In addition, a per-command initialization command
        has to be created and sourced by the shell. Look at argcomplete
        documentation for details.
    """

    def parse(self, context):
        """
        Optionally trigger argument completion in the invoking shell.

        This method is called to see if bash argument completion is requested
        and to honor the request, if needed. This causes the process to exit
        (early) without giving other ingredients a chance to initialize or shut
        down.

        Due to the way argcomple works, no other ingredient can print()
        anything to stdout prior to this point.
        """
        try:
            import argcomplete
        except ImportError:
            return
        try:
            parser = context.parser
        except AttributeError:
            raise RecipeError(
                """
                The context doesn't have the parser attribute.

                The auto-complete ingredient depends on having a parser object
                to generate completion data for she shell.  In a typical
                application this requires that the AutocompleteIngredient and
                ParserIngredient are present and that the auto-complete
                ingredient precedes the parser.
                """)
        else:
            argcomplete.autocomplete(parser)



# encoding: utf-8
# This file is part of Guacamole.
#
# Copyright 2012-2015 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Guacamole is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3,
# as published by the Free Software Foundation.
#
# Guacamole is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Guacamole.  If not, see <http://www.gnu.org/licenses/>.

"""Ingredient for arranging commands into a tree structure."""

from __future__ import absolute_import, print_function, unicode_literals

import collections
import logging
import types

from guacamole.core import Ingredient


_logger = logging.getLogger("guacamole")

#: A named tuple for representing the hierarchy of commands.
#
# The ``cmd_obj`` field is an instance of a Command class.
# The ``cmd_name`` field is the effective name of the command.
# The ``children`` field is a tuple of cmd_tree_node tuples.
cmd_tree_node = collections.namedtuple(
    'cmd_tree_node', 'cmd_name cmd_obj children')


class CommandTreeBuilder(Ingredient):

    """
    Ingredient for arranging commands into a tree of instances.

    Since commands and sub-commands are specified as classes there has to be an
    ingredient that instantiates them and resolves all the naming ambiguities.
    Here it is.

    This component acts early, in its :meth:`added()` method.
    """

    def __init__(self, command):
        """
        Initialize the ingredient with a given top-level command.

        :param command:
            The command that is the top of a commad hierarchy.
        """
        self.command = command

    def added(self, context):
        """
        Ingredient method called before anything else.

        Here this method just builds the full command tree and stores it inside
        the context as the ``cmd_tree`` attribute. The structure of the tree is
        explained by the :func:`build_cmd_tree()` function.
        """
        context.cmd_tree = self._build_cmd_tree(self.command)
        context.cmd_toplevel = context.cmd_tree.cmd_obj
        # Collect spices from the top-level command
        for spice in context.cmd_toplevel.get_cmd_spices():
            context.bowl.add_spice(spice)

    def _build_cmd_tree(self, cmd_cls, cmd_name=None):
        """
        Build a tree of commands.

        :param cmd_cls:
            The Command class or object to start with.
        :param cmd_name:
            Hard-coded name of the command (can be None for auto-detection)
        :returns:
            A tree structure represented as tuple
                ``(cmd_obj, cmd_name, children)``
            Where ``cmd_obj`` is a Command instance, cmd_name is its name, if
            any (it might be None) and ``children`` is a tuple of identical
            tuples.

        Note that command name auto-detection relies on
        :meth:`guacamole.recipes.cmd.Command.get_cmd_name()`.

        Let's look at a simple git-like example::

            >>> from guacamole import Command

            >>> class git_log(Command):
            >>>     pass

            >>> class git_stash_list(Command):
            >>>     pass

            >>> class git_stash(Command):
            >>>     sub_commands = (('list', git_stash_list),)

            >>> class git(Command):
            >>>     sub_commands = (('log', git_log),
            >>>                     ('stash', git_stash))

            >>> build_cmd_tree(git)
            (None, '<git>', (
                ('log', <git_log>, ()),
                ('stash', <git_stash>, (
                    ('list', <git_stash_list>, ()),),),),)
        """
        if isinstance(cmd_cls, type):
            cmd_obj = cmd_cls()
        else:
            cmd_obj = cmd_cls
        if cmd_name is None:
            cmd_name = cmd_obj.get_cmd_name()
        return cmd_tree_node(cmd_name, cmd_obj, tuple([
            self._build_cmd_tree(subcmd_cls, subcmd_name)
            for subcmd_name, subcmd_cls in cmd_obj.get_sub_commands()]))


class CommandTreeDispatcher(Ingredient):

    """
    Ingredient for dispatching commands hierarchically.

    This ingredient builds on the :class:`CommandTreeBuilder` ingredient. It
    implements the :meth:`dispatch()` method that recurses from the top (root)
    of the command tree down to the appropriate leaf, calling the invoke()
    method of each command.

    The process stops on the first command that returns a value other than
    None, raises an exception or until a leaf command is reached. THe ability
    to return early allows commands to perform some sanity checks or short-
    circuit execution that is hard to express using standard parser APIs.

    Lastly, a command can return a generator, this is treated as a sign that
    the generator implements a context-manager-like API. In this case the
    generator is called exactly twice and can be used to manage resources
    during the lifetime of all sub-commands.
    """

    def dispatch(self, context):
        """Dispatch execution to the invoke() method of selected commands."""
        return self._dispatch(context, 0)

    def _dispatch(self, context, level):
        # Find the command we're about to execute.
        try:
            command = getattr(context.args, 'command{}'.format(level))
        except AttributeError:
            return
        else:
            from guacamole.recipes.cmd import Command
            assert isinstance(command, Command)
        # Invoke the command we found, if any.
        _logger.debug("Invoking command %r", command)
        retval = command.invoked(context)
        # Interpret the return value to know what to do next
        if isinstance(retval, types.GeneratorType):
            # Generators are invoked "around" sub-commands.
            # This allows them to use context managers and prepare
            # the execution environment for sub-commands reliably.
            _logger.debug("Command %r uses generator-based invoke. "
                          "Invoking sub-commands (if any)", command)
            return self._dispatch_generator(context, level, retval, command)
        elif retval is None:
            # None is simply ignored and execution continues until a leaf
            # command is reached or until ...
            _logger.debug("Command %r returned None from invoke. "
                          "Invoking sub-commands (if any)", command)
            return self._dispatch_None(context, level, retval, command)
        else:
            # ... or until a non-None result is produced.
            _logger.debug("Command %r returned code %s, returning",
                          command, retval)
            return self._dispatch_other(context, level, retval, command)

    def _dispatch_generator(self, context, level, retval, command):
        # Generators are dispatched with two next() calls.
        # The first one is just there to start executing the code and reach
        # the (first and only) yield statement.
        next(retval)
        # Next, we dispatch the next sub-command.
        try:
            return self._dispatch(context, level + 1)
        finally:
            # Lastly, and this is done in a finally block to ensure it happens
            # in spite of exceptions being thrown. We call the generator again.
            try:
                next(retval)
            except StopIteration:
                pass
            else:
                _logger.error(
                    "BUG in %s.invoke(). "
                    "Each generator-based invoke() MUST use exactly "
                    "one yield statement.", command.__class__.__name__)

    def _dispatch_None(self, context, level, retval, command):
        return self._dispatch(context, level + 1)

    def _dispatch_other(self, context, level, retval, command):
        return retval



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of Guacamole.
#
# Copyright 2012-2015 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Guacamole is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3,
# as published by the Free Software Foundation.
#
# Guacamole is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Guacamole.  If not, see <http://www.gnu.org/licenses/>.
"""setup for guacamole."""

import sys
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')


setup(
    name='guacamole',
    version='0.9.2',
    description='Guacamole is an command line tool library for Python',
    long_description=readme + '\n\n' + history,
    author='Zygmunt Krynicki',
    author_email='me@zygoon.pl',
    url='https://github.com/zyga/guacamole',
    packages=['guacamole', 'guacamole.ingredients', 'guacamole.recipes'],
    package_dir={'guacamole': 'guacamole'},
    include_package_data=True,
    license="LGPLv3",
    zip_safe=True,
    keywords='argparse cli tool command sub-command subcommand',
    tests_require=([
        # XXX: we don't depend on funcsigs but apparently mocks does without
        # properly declaring it. When I run ./setup.py test on Python 2.7 I get
        # an import error on mock, failing to import funcsigs. Oddly enough
        # explicitly installing mock installs funcsigs.
        #
        # If this doesn't happen later, feel free to remove funcsigs.
        'funcsigs',
        'mock',
        'unittest2' if sys.version_info[0] == 2 else 'unittest2py3k',
    ] if sys.version_info[:2] <= (3, 3) else None),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Environment :: Console',
        ('License :: OSI Approved :: GNU Lesser General Public License v3'
         ' (LGPLv3)'),
        'Natural Language :: English',
        'Natural Language :: Polish',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows :: Windows 7',
        'Operating System :: Microsoft :: Windows :: Windows XP',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    test_suite='guacamole',
)
